@@ -2,8 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011 Massachusetts Institute of
-    Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018, 2019 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -78,10 +78,10 @@ USA.
 
 (define (make-decoded-time second minute hour day month year #!optional zone)
   (check-decoded-time-args second minute hour day month year
-			   'MAKE-DECODED-TIME)
+			   'make-decoded-time)
   (let ((zone (if (default-object? zone) #f zone)))
     (if zone
-	(guarantee-time-zone zone 'MAKE-DECODED-TIME))
+	(guarantee time-zone? zone 'make-decoded-time))
     (if zone
 	(%make-decoded-time second minute hour day month year
 			    (compute-day-of-week day month year)
@@ -106,10 +106,10 @@ USA.
 (define (check-decoded-time-args second minute hour day month year caller)
   (let ((check-range
 	 (lambda (object min max)
-	   (guarantee-exact-nonnegative-integer object caller)
+	   (guarantee exact-nonnegative-integer? object caller)
 	   (if (not (<= min object max))
 	       (error:bad-range-argument object caller)))))
-    (guarantee-exact-nonnegative-integer year caller)
+    (guarantee exact-nonnegative-integer? year caller)
     (check-range month 1 12)
     (check-range day 1 (month/max-days month))
     (check-range hour 0 23)
@@ -166,6 +166,9 @@ USA.
 
 (define epoch 2208988800)
 
+(define (current-second)
+  (exact->inexact ((ucode-primitive encoded-time 0))))
+
 (define (local-decoded-time)
   (universal-time->local-decoded-time (get-universal-time)))
 
@@ -273,7 +276,7 @@ USA.
 (define (rfc2822-string->decoded-time string)
   (let ((v (*parse-string parser:rfc2822-time string)))
     (if (not v)
-	(error:bad-range-argument string 'STRING->DECODED-TIME))
+	(error:bad-range-argument string 'string->decoded-time))
     (vector-ref v 0)))
 
 (define (string->universal-time string)
@@ -349,7 +352,7 @@ USA.
       (write-time-zone tz port))))
 
 (define (write-time-zone tz port)
-  (guarantee-time-zone tz 'WRITE-TIME-ZONE)
+  (guarantee time-zone? tz 'write-time-zone)
   (let ((minutes (round (* 60 (- tz)))))
     (let ((qr (integer-divide (abs minutes) 60)))
       (write-char (if (< minutes 0) #\- #\+) port)
@@ -359,7 +362,7 @@ USA.
 (define (string->time-zone string)
   (let ((v (*parse-string parser:time-zone string)))
     (if (not v)
-	(error:bad-range-argument string 'STRING->TIME-ZONE))
+	(error:bad-range-argument string 'string->time-zone))
     (vector-ref v 0)))
 
 (define parser:time-zone
@@ -458,7 +461,7 @@ USA.
 	 (*parse-string (parser:ctime (if (default-object? zone) #f zone))
 			string)))
     (if (not v)
-	(error:bad-range-argument string 'CTIME-STRING->DECODED-TIME))
+	(error:bad-range-argument string 'ctime-string->decoded-time))
     (vector-ref v 0)))
 
 (define (universal-time->local-ctime-string time)
@@ -481,7 +484,7 @@ USA.
 
 (define (parser:ctime zone)
   (if zone
-      (guarantee-time-zone zone 'PARSER:CTIME))
+      (guarantee time-zone? zone 'parser:ctime))
   (*parser
    (encapsulate (lambda (v)
 		  (make-decoded-time (vector-ref v 5)
@@ -524,14 +527,14 @@ USA.
 (define (iso8601-string->decoded-time string #!optional start end)
   (let ((v (*parse-string parser:iso8601-date/time string start end)))
     (if (not v)
-	(error:bad-range-argument string 'ISO8601-STRING->DECODED-TIME))
+	(error:bad-range-argument string 'iso8601-string->decoded-time))
     (vector-ref v 0)))
 
 (define (xml-rpc-iso8601-string->decoded-time string #!optional start end)
   (let ((v (*parse-string parser:xml-rpc-iso8601-date/time string start end)))
     (if (not v)
 	(error:bad-range-argument string
-				  'XML-RPC-ISO8601-STRING->DECODED-TIME))
+				  'xml-rpc-iso8601-string->decoded-time))
     (vector-ref v 0)))
 
 (define (decoded-time->iso8601-string dt)
@@ -870,15 +873,15 @@ USA.
 ;;;; Utilities
 
 (define (month/max-days month)
-  (guarantee-month month 'MONTH/MAX-DAYS)
+  (guarantee-month month 'month/max-days)
   (vector-ref '#(31 29 31 30 31 30 31 31 30 31 30 31) (- month 1)))
 
 (define (month/short-string month)
-  (guarantee-month month 'MONTH/SHORT-STRING)
+  (guarantee-month month 'month/short-string)
   (vector-ref month/short-strings (- month 1)))
 
 (define (month/long-string month)
-  (guarantee-month month 'MONTH/LONG-STRING)
+  (guarantee-month month 'month/long-string)
   (vector-ref month/long-strings (- month 1)))
 
 (define (guarantee-month month name)
@@ -894,11 +897,11 @@ USA.
 	     (error "Unknown month designation:" string))))
 
 (define (day-of-week/short-string day)
-  (guarantee-day-of-week day 'DAY-OF-WEEK/SHORT-STRING)
+  (guarantee-day-of-week day 'day-of-week/short-string)
   (vector-ref days-of-week/short-strings day))
 
 (define (day-of-week/long-string day)
-  (guarantee-day-of-week day 'DAY-OF-WEEK/LONG-STRING)
+  (guarantee-day-of-week day 'day-of-week/long-string)
   (vector-ref days-of-week/long-strings day))
 
 (define (guarantee-day-of-week day name)
@@ -922,7 +925,7 @@ USA.
 (define (string->year string)
   (let ((n (string->number string)))
     (if (not (exact-nonnegative-integer? n))
-	(error:bad-range-argument string 'STRING->YEAR))
+	(error:bad-range-argument string 'string->year))
     (cond ((< n 70) (+ 2000 n))
 	  ((< n 100) (+ 1900 n))
 	  (else n))))
